//
//  OMAutomaton.h
//
//  Created by Samuel Grau on 1/23/12.
//  Copyright (c) 2012 Samuel GRAU. All rights reserved.
//

#import <Foundation/Foundation.h>

@class OMAState;

typedef enum {
	/**
	 * Minimize using Huffman's O(n<sup>2</sup>) algorithm. 
	 * This is the standard text-book algorithm.
	 * @see #setMinimization(int)
	 */
	OMAMinimizationAlgorithmHuffman = 0,
    
    /**
	 * Minimize using Brzozowski's O(2<sup>n</sup>) algorithm. 
	 * This algorithm uses the reverse-determinize-reverse-determinize trick, which has a bad
	 * worst-case behavior but often works very well in practice 
	 * (even better than Hopcroft's!).
	 * @see #setMinimization(int)
	 */
    OMAMinimizationAlgorithmTypeBrozozowski = 1,
    
    /**
	 * Minimize using Hopcroft's O(n log n) algorithm.
	 * This is regarded as one of the most generally efficient algorithms that exist.
	 * @see #setMinimization(int)
	 */
	OMAMinimizationAlgorithmTypeHopcroft = 2
    
} OMAMinimizationAlgorithmType;

/**
 * Finite-state automaton with regular expression operations.
 * <p>
 * Class invariants:
 * <ul>
 * <li> An automaton is either represented explicitly (with {@link OMAState} and {@link OMATransition} objects)
 *      or with a singleton string (see {@link #getSingleton()} and {@link #expandSingleton()}) 
 *      in case the automaton is known to accept exactly one string.
 *      (Implicitly, all states and transitions of an automaton are reachable from its initial state.)
 * <li> Automata are always reduced (see {@link #reduce()}) 
 *      and have no transitions to dead states (see {@link #removeDeadTransitions()}).
 * <li> If an automaton is nondeterministic, then {@link #isDeterministic()} returns false (but
 *      the converse is not required).
 * <li> Automata provided as input to operations are generally assumed to be disjoint.
 * </ul>
 * <p>
 * If the states or transitions are manipulated manually, the {@link #restoreInvariant()}
 * and {@link #setDeterministic(boolean)} methods should be used afterwards to restore 
 * representation invariants that are assumed by the built-in automata operations.
 * 
 */
@interface OMAutomaton : NSObject {
    OMAMinimizationAlgorithmType _minimizationAlgorithm;
    OMAState * _initialState;
    BOOL _isDeterministic;
    BOOL _alwaysMinimize;
    BOOL _allowMutation;
}

/** Initial state of this automaton. */
@property (nonatomic, readonly) OMAState * initialState;

/**
 * Returns deterministic flag for this automaton.
 * @return true if the automaton is definitely deterministic, false if the automaton
 *         may be nondeterministic
 */
@property (nonatomic, readonly) BOOL isDeterministic;

/** 
 * Selects minimization algorithm (default: <code>OMAMinimizationAlgorithmTypeHopcroft</code>). 
 * @param algorithm minimization algorithm
 */
- (void)setMinimization:(OMAMinimizationAlgorithmType)minimizationType;

/**
 * Sets or resets minimize always flag.
 * If this flag is set, then {@link #minimize()} will automatically
 * be invoked after all operations that otherwise may produce non-minimal automata.
 * By default, the flag is not set.
 * @param flag if true, the flag is set
 */
- (void)setAlwaysMinimizeFlag:(BOOL)flag;

/**
 * Sets or resets allow mutate flag.
 * If this flag is set, then all automata operations may modify automata given as input;
 * otherwise, operations will always leave input automata languages unmodified. 
 * By default, the flag is not set.
 * @param flag if true, the flag is set
 * @return previous value of the flag
 */
- (BOOL)setAllowMutation:(BOOL)allowMutation;

/**
 * Returns the state of the allow mutate flag.
 * If this flag is set, then all automata operations may modify automata given as input;
 * otherwise, operations will always leave input automata languages unmodified. 
 * By default, the flag is not set.
 * @return current value of the flag
 */
- (BOOL)allowMutation;

/** 
 * Returns the set of states that are reachable from the initial state.
 * @return set of {@link OMAState} objects
 */
- (NSMutableSet *)states;

/** 
 * Returns the set of reachable accept states. 
 * @return set of {@link State} objects
 */
- (NSMutableSet *)acceptStates;

/**
 * Restores representation invariant.
 * This method must be invoked before any built-in automata operation is performed 
 * if automaton states or transitions are manipulated manually.
 * @see #setDeterministic(boolean)
 */
- (void)restoreInvariant;

/** 
 * Removes transitions to dead states and calls {@link #reduce()} and {@link #clearHashCode()}.
 * (A state is "dead" if no accept state is reachable from it.)
 */
- (void)removeDeadTransitions;

/** 
 * Returns the set of live states. A state is "live" if an accept state is reachable from it. 
 * @return set of {@link OMAState} objects
 */
- (NSMutableSet *)liveStates;

/**
 * Returns the number of states in this automaton.
 * @return the number of states of the automaton
 */
- (NSUInteger)numberOfStates;

/**
 * Returns the number of transitions in this automaton. This number is counted
 * as the total number of edges, where one edge may be a character interval.
 */
- (NSUInteger)numberOfTransitions;

- (void)minimize;

/** 
 * Reduces this automaton.
 * An automaton is "reduced" by combining overlapping and adjacent edge intervals 
 * with same destination. 
 */
- (void)reduce;

@end
