//
//  Automaton.m
//
//  Created by Samuel Grau on 1/23/12.
//  Copyright (c) 2012 Samuel GRAU. All rights reserved.
//

#import "OMAutomaton.h"
#import "OMAState.h"
#import "OMATransition.h"


@implementation OMAutomaton

@synthesize initialState = _initialState;
@synthesize isDeterministic = _isDeterministic;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#pragma mark -
#pragma mark Initialization
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
- (id)init {
    self = [super init];
    if (self) {
        [self setMinimization:OMAMinimizationAlgorithmTypeHopcroft];
        [self setAlwaysMinimizeFlag:NO];
        self->_allowMutation = NO;
        self->_isDeterministic = YES;
        
        self->_initialState = [[OMAState alloc] init];
    }
    return self;
}



// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#pragma mark -
#pragma mark Memory Management
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
- (void)dealloc {
    ANReleaseSafely(&self->_initialState);
    [super dealloc];
}



// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#pragma mark -
#pragma mark Public
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
- (void)setMinimization:(OMAMinimizationAlgorithmType)minimizationType {
    self->_minimizationAlgorithm = minimizationType;
}


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
- (void)setAlwaysMinimizeFlag:(BOOL)flag {
    self->_alwaysMinimize = flag;
}


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
- (BOOL)setAllowMutation:(BOOL)allowMutation {
    BOOL b = self->_allowMutation;
    self->_allowMutation = allowMutation;
    return b;
}


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
- (BOOL)allowMutation {
    return self->_allowMutation;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
- (void)checkMinimizeAlways {
    if (self->_alwaysMinimize) {
        [self minimize];
    }
}


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
- (NSMutableSet *)states {
    OMA_NILinkedList * workList = [[OMA_NILinkedList alloc] init];
    NSMutableSet * visited = [[NSMutableSet alloc] init];
    
    [workList addObject:[self initialState]];
    [visited addObject:[self initialState]];
    
    while ([workList count]) {
        OMAState * s = [workList firstObject];
        [workList removeFirstObject];
        
        for (OMATransition * t in [s transitions]) {
            if (![visited containsObject:[t to]]) {
                [visited addObject:[t to]];
                [workList addObject:[t to]];
            }
        }
    }
    
    ANReleaseSafely(&workList);
    
    return [visited autorelease];
}


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
- (NSMutableSet *)acceptStates {
    NSMutableSet * accepts = [[NSMutableSet alloc] init];
    NSMutableSet * visited = [[NSMutableSet alloc] init];
    OMA_NILinkedList * workList = [[OMA_NILinkedList alloc] init];
    
    [workList addObject:[self initialState]];
    [visited addObject:[self initialState]];
    
    while ([workList count]) {
        OMAState * s = [workList firstObject];
        [workList removeFirstObject];
        
        if ([s isAcceptState]) {
            [accepts addObject:s];
        }
        
        for (OMATransition * t in [s transitions]) {
            if (![visited containsObject:[t to]]) {
                [visited addObject:[t to]];
                [workList addObject:[t to]];
            }
        }
    }
    
    ANReleaseSafely(&workList);
    ANReleaseSafely(&visited);
    
    return [accepts autorelease];
}


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
- (void)restoreInvariant {
    [self removeDeadTransitions];
}


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
- (void)removeDeadTransitions {
    NSMutableSet * states = [self states];
    NSMutableSet * live = [self liveStates];
    
    //Set<State> live = getLiveStates(states);
    for (OMAState * s in states) {
        NSMutableSet * st = [s transitions];
        [s resetTransitions];
        for (OMATransition * t in st) {
            if ([live containsObject:[t to]]) {
                [st addObject:t];
            }
        }
    }
    [self reduce];
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
- (NSMutableSet *)liveStates:(NSMutableSet *)states {
    NSMutableDictionary * map = [[NSMutableDictionary alloc] init];
    for (OMAState * s in states) {
        NSMutableSet * set = [[NSMutableSet alloc] init];
        [map setObject:set forKey:s];
        ANReleaseSafely(&set);
    }
    
    for (OMAState * s in states) {
        for (OMATransition * t in [s transitions]) {
            NSMutableSet * set = [map objectForKey:[t to]];
            [set addObject:t];
        }
    }
    
    NSMutableSet * live = [[NSMutableSet alloc] init];
    [live addObjectsFromArray:[[self acceptStates] allObjects]];

    OMA_NILinkedList * workList = [[OMA_NILinkedList alloc] initWithArray:[live allObjects]];
    
    while ([workList count]) {
        OMAState * s = [workList firstObject];
        [workList removeFirstObject];
        
        for (OMAState * p in [map objectForKey:s]) {
            if (![live containsObject:p]) {
                [live addObject:p];
                [workList addObject:p];
            }
        }
    }
    
    ANReleaseSafely(&workList);
    
    return [live autorelease];
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
- (NSMutableSet *)liveStates {
    return [self liveStates:[self states]];
}


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
- (NSUInteger)numberOfStates {
    return [[self states] count];
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
- (NSUInteger)numberOfTransitions {
    NSUInteger c = 0;
    NSMutableSet * states = [self states];
    for (OMAState * s in states) {
        c += [[s transitions] count];
    }
    return c;
}


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
- (void)minimize {
    // TODO
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
- (void)reduce {
    NSMutableSet * states = [self states];
    
    //setStateNumbers(states);
    for (OMAState * s in states) {
        
        NSMutableSet * st = [s transitions];
        
        [s resetTransitions];
        OMAState * p = nil;
        
        
        for (OMATransition * t in st) {
            if (p == [t to]) {
                OMATransition * t = [[OMATransition alloc] initWithDestinationState:p 
                                                                           andValue:t.value];
                [[s transitions] addObject:t];
                ANReleaseSafely(&t);
                
            } else {
                if (p != nil) {
                    OMATransition * t = [[OMATransition alloc] initWithDestinationState:p 
                                                                               andValue:t.value];
                    [[s transitions] addObject:t];
                    ANReleaseSafely(&t);
                }
                p = t.to;
            }
        }
        
        if (p != nil) {
            OMATransition * t = [[OMATransition alloc] initWithDestinationState:p 
                                                                       andValue:t.value];
            [[s transitions] addObject:t];
            ANReleaseSafely(&t);
        }
    }
    
    //clearHashCode();
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


@end
